#define __STDC_FORMAT_MACROS
#include "ffdb.h"
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <new>

//When the buffer chances location, adjust the pointers for the fields
void Row::adjustptrs(void *oldptr_, void *newptr_) {
	char *oldptr=(char*)oldptr_;
	char *newptr=(char*)newptr_;
	for(container_type::iterator iter=col_ptr.begin();
	    iter!=col_ptr.end();
	    iter++)
	{
		char *o = (char*)iter->ptr;
		if(o) {
			ptrdiff_t offset = o-oldptr;
			iter->ptr = newptr + offset;
		}
	}
}


Row::Row()
  : col_ptr(),
    buffer_bytes(0),
    buffer_size(0),
    buffer(0)
{
}


Row::Row(const Row &row_)
  : col_ptr(),
    buffer_bytes(0),
    buffer_size(0),
    buffer(0)
{
	*this = row_;
}


Row& Row::operator=(const Row &row_) {
	if(&row_==this) return *this;
	
	void *new_buffer;
	if(row_.buffer) {
		new_buffer = malloc(row_.buffer_bytes);
		if(!new_buffer) throw std::bad_alloc();
	} else
		new_buffer = 0;
	try {
		col_ptr = row_.col_ptr;
	} catch(std::bad_alloc&) {
		free(new_buffer);
		throw;
	}
	free(buffer);
	buffer = new_buffer;
	
	if(row_.buffer) {
		memcpy(buffer,row_.buffer,row_.buffer_bytes);
		buffer_bytes = row_.buffer_bytes;
		buffer_size = row_.buffer_bytes;
		adjustptrs(row_.buffer,buffer);
	} else {
		buffer_bytes=0;
		buffer_size=0;
		buffer=0;
	}
	return *this;
}


void Row::reserve_raw(size_t bytes) {
	if(bytes>buffer_size) {
		if(bytes<8) {
			bytes = (bytes+3)&~3;
		} else if(bytes<16) {
			bytes = (bytes+7)&~7;
		} else if(bytes<256) {
			bytes = (bytes+15)&~15;
		} else
			bytes = (bytes+63)&~63;
		
		char *newbuf = (char*)realloc(buffer,bytes);
		if(!newbuf) throw std::bad_alloc();
		adjustptrs(buffer,newbuf);
		buffer = newbuf;
		buffer_size = bytes;
	}
}




void Row::resize_columns(size_t columns) {
	col_ptr.resize(columns,VL(0,0));
}


void *Row::alloc(size_t bytes, size_t alignment) {
	size_t start_offset = buffer_bytes;
	while((start_offset)%alignment)
		start_offset++;
	
	size_t endoffset = start_offset + bytes;
	reserve_raw(endoffset);
	
	buffer_bytes = endoffset;
	void *ptr = (char*)buffer + start_offset;
	col_ptr.push_back(VL(ptr,bytes));
	return ptr;
}


void *Row::alloc(size_t bytes, size_t alignment, const void *value) {
	if(value) {
		void *ptr = alloc(bytes,alignment);
		memcpy(ptr,value,bytes);
		return ptr;
	} else {
		col_ptr.push_back(VL(0,0));
		return 0;
	}
}

void Row::clear() {
	free(buffer);
	buffer=0;
	buffer_bytes=0;
	buffer_size=0;
	col_ptr.clear();
}


bool Row::operator==(const Row &row2) const {
	if(!col_ptr.size()!=row2.col_ptr.size()) return false;
	for(size_type i=0; i<col_ptr.size(); i++) {
		const VL &vl1 = col_ptr[i];
		const VL &vl2 = row2.col_ptr[i];
		if(vl1.size!=vl2.size) return false;
		if(vl1.ptr==vl2.ptr) continue; //handles NULL
		if(memcmp(vl1.ptr,vl2.ptr,vl1.size)!=0) return false;
	}
	return true;
}


bool Row::operator!=(const Row &row2) const {
	if(!col_ptr.size()==row2.col_ptr.size()) return false;
	for(size_type i=0; i<col_ptr.size(); i++) {
		const VL &vl1 = col_ptr[i];
		const VL &vl2 = row2.col_ptr[i];
		if(vl1.size!=vl2.size) continue;
		if(vl1.ptr==vl2.ptr) return false; //handles NULL
		if(memcmp(vl1.ptr,vl2.ptr,vl1.size)==0) return false;
	}
	return true;
}


void Row::swap(Row &row2) {
	std::swap(col_ptr,row2.col_ptr);
	std::swap(buffer_bytes,row2.buffer_bytes);
	std::swap(buffer_size,row2.buffer_size);
	std::swap(buffer,row2.buffer);
}





bool is_type_compatible(const RowSetDescriptor &desc1, const RowSetDescriptor &desc2) {
	if(desc1.size()!=desc2.size())
		return false;
	for(unsigned i=0; i<desc1.size(); i++) {
		if(desc1[i].type != desc2[i].type)
			return false;
	}
	return true;
}





RowSet::~RowSet() {
}
	
void RowSet::setDescriptor(const RowSetDescriptor &new_desc) {
	desc = new_desc;
	row.clear();
}

void RowSet::push_back(const Row &new_row) {
	row.push_back(new_row);
}


void RowSet::add_and_swap(Row &new_row) {
	row.push_back(Row());
	iterator i = row.end()-1;
	i->swap(new_row);
}
