#define __STDC_FORMAT_MACROS
#include "ffdb_convert.h"
#include "ffdb_op_alignment.h"
#include <stdio.h>

bool LockedRowConverter::prepare_cache()
{
	if(cache_valid) return true;
	
	cache_valid = false;
	
	for(std::vector<Conversion>::iterator iter=conv.begin();
	    iter!=conv.end();
	    iter++)
	{
		unsigned column_idx;
		for(column_idx=0; column_idx<desc.size(); column_idx++) {
			if(desc[column_idx].name==iter->column_name) break;
		}
		if(column_idx>=desc.size()) return false;
		iter->column_idx = column_idx;
		iter->column_type = desc[column_idx].type;
	}
	
	desc_conv_idx.clear();
	for(unsigned column_idx=0; column_idx<desc.size(); column_idx++) {
		int conv_idx=0;
		std::vector<Conversion>::iterator iter=conv.begin();
		for(;
		    iter!=conv.end();
		    iter++)
		{
			if(desc[column_idx].name==iter->column_name) break;
			conv_idx++;
		}
		if(iter!=conv.end())
			desc_conv_idx.push_back(conv_idx);
		else
			desc_conv_idx.push_back(-1);
	}
	
	cache_valid=true;
	return true;
}


void LockedRowConverter::describe_(const char *column_name, native_type_t native_type, void *val, size_t maxsize, bool *null_indicator)
{
	Conversion c;
	c.column_name=column_name;
	c.native_type = native_type;
	c.native_val = val;
	c.maxsize = maxsize;
	c.null_indicator = null_indicator;
	conv.push_back(c);
}


LockedRowConverter::~LockedRowConverter()
{
}


void LockedRowConverter::lock(const RowSetDescriptor &desc_)
{
	if(desc!=desc_) {
		cache_valid=false;
		desc=desc_;
	}
}


bool LockedRowConverter::toRow(Row *row) {
	if(!prepare_cache()) return false;
	row->clear();
	for(std::vector<int>::iterator iter=desc_conv_idx.begin();
	    iter!=desc_conv_idx.end();
	    iter++)
	{
		if(*iter!=-1) {
			if(!toColumn(conv[*iter],row))
				return false;
		} else
			row->alloc(0,0,0);
	}
	return true;
}


bool LockedRowConverter::toNative(const Row &row) {
	if(!prepare_cache()) return false;
	for(std::vector<Conversion>::iterator iter=conv.begin();
	    iter!=conv.end();
	    iter++)
	{
		if(!toNative(row[iter->column_idx],*iter))
			return false;
	}
	return true;
}


bool LockedRowConverter::toColumn(const Conversion &c, Row *row) {
	if(c.native_val==0 || (c.null_indicator && *c.null_indicator)) {
		row->alloc(0,0,0);
		return true;
	}
	
	switch(c.column_type) {
		case ColumnType_integer: {
			int64_t value;
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: value = (int64_t)*(const t*)c.native_val; if(value!=*(const t*)c.native_val) return false; break;
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char: {
					if(sscanf((const char*)c.native_val, "%"SCNd64, &value)!=1)
						return false;
					break;
				}
				case nt_string: {
					if(sscanf(((const std::string*)c.native_val)->c_str(), "%"SCNd64, &value)!=1)
						return false;
					break;
				}
			}
			row->alloc(sizeof(value),ffdb_type_alignment[ColumnType_integer],&value);
			break;
		}
		case ColumnType_real: {
			double value;
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: value = *(const t*)c.native_val; if(value!=*(const t*)c.native_val) return false; break;
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char: {
					if(sscanf((const char*)c.native_val, "%le", &value)!=1)
						return false;
					break;
				}
				case nt_string: {
					if(sscanf(((const std::string*)c.native_val)->c_str(), "%le", &value)!=1)
						return false;
					break;
				}
			}
			row->alloc(sizeof(value),ffdb_type_alignment[ColumnType_real],&value);
			break;
		}
		case ColumnType_string: {
			char value[64];
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,f,t) case e: sprintf(value,"%"f, *(const t*)c.native_val); break;
				HANDLE(nt_s16,PRId16,int16_t)
				HANDLE(nt_u16,PRIu16,uint16_t)
				HANDLE(nt_s32,PRId32,int32_t)
				HANDLE(nt_u32,PRIu32,uint32_t)
				HANDLE(nt_s64,PRId64,int64_t)
				HANDLE(nt_u64,PRIu64,uint64_t)
				HANDLE(nt_float,"e",float)
				HANDLE(nt_double,"e",double)
				case nt_char: {
					size_t sl = strlen((const char*)c.native_val);
					row->alloc(sl+1,ffdb_type_alignment[ColumnType_string],c.native_val);
					return true;
				}
				case nt_bool: {
					if(*(const bool*)c.native_val)
						row->alloc(2,ffdb_type_alignment[ColumnType_string],"T");
					else
						row->alloc(2,ffdb_type_alignment[ColumnType_string],"F");
					return true;
				}
				case nt_string: {
					const std::string *s = (const std::string*)c.native_val;
					void *p = row->alloc(s->size()+1,ffdb_type_alignment[ColumnType_string],0);
					memcpy(p,s->data(),s->size());
					((char*)p)[s->size()]='\0';
					return true;
				}
			}
			row->alloc(strlen(value)+1,ffdb_type_alignment[ColumnType_string],value);
			break;
		}
		case ColumnType_raw: {
			switch(c.native_type) {
				case nt_char: {
					size_t sl = strlen((const char*)c.native_val);
					row->alloc(sl+1, ffdb_type_alignment[ColumnType_raw], c.native_val);
					break;
				}
				case nt_string: {
					const std::string *s = (const std::string*)c.native_val;
					row->alloc(s->size(), ffdb_type_alignment[ColumnType_raw], s->data());
					break;
				}
				default:
					row->alloc(c.maxsize, ffdb_type_alignment[ColumnType_raw], c.native_val);
					break;
			}
			break;
		}
		case ColumnType_boolean: {
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: { bool b = *(const t*)c.native_val; row->alloc(sizeof(b),ffdb_type_alignment[ColumnType_boolean],&b); break; }
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char: {
					const char *s = (const char*)c.native_val;
					bool b;
					if(strcmp(s,"T")==0 || strcmp(s,"true")==0)
						b = true;
					else if(strcmp(s,"F")==0 || strcmp(s,"false")==0)
						b = false;
					else
						return false; //cannot convert
					row->alloc(sizeof(b),ffdb_type_alignment[ColumnType_boolean],&b);
					break;
				}
				case nt_string: {
					const std::string *s = (const std::string*)c.native_val;
					bool b;
					if(*s=="T" || *s=="true")
						b = true;
					else if(*s=="F" || *s=="false")
						b = false;
					else
						return false; //cannot convert
					row->alloc(sizeof(b),ffdb_type_alignment[ColumnType_boolean],&b);
					break;
				}
			}
			break;
		}
		default:
			return false;
	}
	return true;
}


bool LockedRowConverter::toNative(const Row::VL &vl, Conversion &c) {
	if(vl.ptr) {
		if(c.null_indicator)
			*c.null_indicator = false;
	} else {
		if(c.null_indicator) {
			*c.null_indicator = true;
			return true;
		} else
			return false;
	}

	switch(c.column_type) {
		case ColumnType_integer: {
			int64_t value = *(const int64_t*)vl.ptr;
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: if((t)value!=value) return false; *(t*)c.native_val = (t)value; break;
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char:
					snprintf((char*)c.native_val,c.maxsize,"%"PRId64, value);
					break;
				case nt_string: {
					char s[22];
					snprintf(s,sizeof(s),"%"PRId64, value);
					*(std::string*)c.native_val = s;
					break;
				}
			}
			break;
		}
		case ColumnType_real: {
			double value = *(const double*)vl.ptr;
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: if((t)value!=value) return false; *(t*)c.native_val = (t)value; break;
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char:
					snprintf((char*)c.native_val,c.maxsize,"%e", value);
					break;
				case nt_string: {
					char s[22];
					snprintf(s,sizeof(s),"%e", value);
					*(std::string*)c.native_val = s;
					break;
				}
			}
			break;
		}
		case ColumnType_string: {
			char s[64];
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,f,t) case e: sscanf(s,"%"f, (t*)c.native_val); break;
				HANDLE(nt_s16,SCNd16,int16_t)
				HANDLE(nt_u16,SCNu16,uint16_t)
				HANDLE(nt_s32,SCNd32,int32_t)
				HANDLE(nt_u32,SCNu32,uint32_t)
				HANDLE(nt_s64,SCNd64,int64_t)
				HANDLE(nt_u64,SCNu64,uint64_t)
				HANDLE(nt_float,"e",float)
				HANDLE(nt_double,"le",double)
				case nt_bool: {
					if(strcmp((const char*)vl.ptr,"T")==0 || strcmp((const char*)vl.ptr,"true")==0)
						*(bool*)c.native_val = true;
					else if(strcmp((const char*)vl.ptr,"F")==0 || strcmp((const char*)vl.ptr,"false")==0)
						*(bool*)c.native_val = false;
					else
						return false; //cannot convert
					break;
				}
				case nt_char: {
					if(vl.size>c.maxsize) return false; //too large
					memcpy(c.native_val, vl.ptr, vl.size);
					break;
				}
				case nt_string: {
					std::string *s = (std::string*)c.native_val;
					s->assign((const char*)vl.ptr,vl.size-1);
					break;
				}
			}
			break;
		}
		case ColumnType_raw: {
			switch(c.native_type) {
				case nt_char: {
					if(vl.size<1) return false;
					if(((const char*)vl.ptr)[vl.size-1]!='\0') return false;
					if(memchr(vl.ptr,'\0',vl.size-1)) return false;
					memcpy(c.native_val, vl.ptr, vl.size);
					break;
				}
				case nt_string: {
					std::string *s = (std::string*)c.native_val;
					s->assign((const char*)vl.ptr,vl.size-1);
					break;
				}
				default: {
					if(vl.size!=c.maxsize) return false;
					memcpy(c.native_val,vl.ptr,vl.size);
					break;
				}
			}
			break;
		}
		case ColumnType_boolean: {
			switch(c.native_type) {
#undef HANDLE
#define HANDLE(e,t) case e: *(t*)c.native_val = *(const bool*)vl.ptr; break;
				HANDLE(nt_s16,int16_t)
				HANDLE(nt_u16,uint16_t)
				HANDLE(nt_s32,int32_t)
				HANDLE(nt_u32,uint32_t)
				HANDLE(nt_s64,int64_t)
				HANDLE(nt_u64,uint64_t)
				HANDLE(nt_float,float)
				HANDLE(nt_double,double)
				HANDLE(nt_bool,bool)
				case nt_char: {
					if(c.maxsize<6) return false;
					if(*(const bool*)vl.ptr)
						strcpy((char*)c.native_val,"true");
					else
						strcpy((char*)c.native_val,"false");
					break;
				}
				case nt_string: {
					if(*(const bool*)vl.ptr)
						*((std::string*)c.native_val) = "true";
					else
						*((std::string*)c.native_val) = "false";
					break;
				}
			}
			break;
		}
		default:
			return false;
	}
	return true;
}


bool RowConverter::toRow(const RowSetDescriptor &desc, Row *row) {
	lock(desc);
	return LockedRowConverter::toRow(row);
}


bool RowConverter::toNative(const RowSetDescriptor &desc, const Row &row) {
	lock(desc);
	return LockedRowConverter::toNative(row);
}
