#ifndef _FFDB_CONVERT_H_
#define _FFDB_CONVERT_H_

#include "ffdb.h"
#include <inttypes.h>
#include <unistd.h>
#include <string>
#include <vector>

class LockedRowConverter {
public:
	enum native_type_t {
		nt_s16, nt_u16,
		nt_s32, nt_u32,
		nt_s64, nt_u64,
		nt_char,
		nt_bool,
		nt_float,
		nt_double,
		nt_string
	};
private:
	struct Conversion {
		std::string column_name;
		unsigned column_idx;
		ColumnType_t column_type;
		native_type_t native_type;
		void *native_val;
		size_t maxsize;
		bool *null_indicator;
	};
	std::vector<Conversion> conv;
	std::vector<int> desc_conv_idx;
	bool cache_valid;
	RowSetDescriptor desc;
	bool prepare_cache();
	void describe_(const char *column_name, native_type_t native_type, void *val, size_t maxsize, bool *null_indicator=0);
	bool toColumn(const Conversion &c, Row *row);
	bool toNative(const Row::VL &vl, Conversion &c);
public:	
	LockedRowConverter() : cache_valid(false) {}
	virtual ~LockedRowConverter();
	
	void lock(const RowSetDescriptor &desc_);
	void lock(const RowSet &rowset) { lock(rowset.descriptor()); }
	
	void describe(const char *column_name, int16_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_s16,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, uint16_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_u16,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, int32_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_s32,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, uint32_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_u32,val,sizeof(*val),null_indicator); }
#if defined(_XBS5_ILP32_OFF32) || defined(_XBS5_ILP32_OFFBIG) || defined(_POSIX_V6_ILP32_OFF32) || defined(_POSIX_V6_ILP32_OFFBIG)
	//long not covered by [u]intXX_t
	void describe(const char *column_name, long *val, bool *null_indicator=0)
	  { describe_(column_name,nt_s32,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, unsigned long *val, bool *null_indicator=0)
	  { describe_(column_name,nt_u32,val,sizeof(*val),null_indicator); }
#endif
	void describe(const char *column_name, int64_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_s64,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, uint64_t *val, bool *null_indicator=0)
	  { describe_(column_name,nt_u64,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, bool *val, bool *null_indicator=0)
	  { describe_(column_name,nt_bool,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, float *val, bool *null_indicator=0)
	  { describe_(column_name,nt_float,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, double *val, bool *null_indicator=0)
	  { describe_(column_name,nt_double,val,sizeof(*val),null_indicator); }
	void describe(const char *column_name, char *val, size_t maxsize, bool *null_indicator=0)
	  { describe_(column_name,nt_char,val,maxsize,null_indicator); }
	void describe(const char *column_name, std::string *val, bool *null_indicator=0)
	  { describe_(column_name,nt_string,val,0,null_indicator); }
	
	bool toRow(Row *row);
	bool toNative(const Row &row);
};


class RowConverter : public LockedRowConverter {
public:
	bool toRow(const RowSetDescriptor &desc, Row *row);
	bool toNative(const RowSetDescriptor &desc, const Row &row);
};

#endif
/*
struct user {
	int id;
	char username[30];
};

describe("id", &user.id);
describe("username", user.username, sizeof(user.username));

*/
