#include "ffdb_op.h"
#include "ffdb_op_cmpdigest.h"
#include <algorithm>
#include <stdarg.h>

#include <stdio.h>



bool LockedDifference::prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2)
{
	if(cache_valid &&
	  row_set_1.descriptor()==cached_desc1 &&
	  row_set_2.descriptor()==cached_desc2)
		return true;
	cache_valid = false;
	
	if(!is_type_compatible(row_set_1.descriptor(),row_set_2.descriptor()))
		return false;
	cached_desc1 = row_set_1.descriptor();
	cached_desc2 = row_set_2.descriptor();
	if(!BinaryOpHelper::prepare_cache(cached_desc1,cached_desc2))
		return false;
	
	cache_valid = true;
	return true;
}


LockedDifference::~LockedDifference() {
}


bool LockedDifference::execute(RowSet *result)
{
	if(!chunk1.rowset || !chunk2.rowset) return false;
	if(!prepare_cache(*chunk1.rowset,*chunk2.rowset))
		return false;
	process_rowset_1(true,false);
	process_rowset_2(true,false);
	
	result->setDescriptor(cached_desc1);
	
	Compare_RowDigest cmp;
	RowSetDigest::const_iterator i = chunk1.rowset_digest.begin();
	RowSetDigest::const_iterator j = chunk2.rowset_digest.begin();
	while(i!=chunk1.rowset_digest.end() && j!=chunk2.rowset_digest.end()) {
		if(cmp.lt(*i,*j)) {
			result->push_back(*i->row_ptr);
			i++;
		} else if(cmp.gt(*i,*j)) {
			result->push_back(*j->row_ptr);
			j++;
		} else {
			i++;
			j++;
		}
	}
	while(i!=chunk1.rowset_digest.end()) {
		result->push_back(*i->row_ptr);
		i++;
	}
	while(j!=chunk2.rowset_digest.end()) {
		result->push_back(*j->row_ptr);
		j++;
	}
	
	return true;
}


bool PreparedDifference::execute(const RowSet &row_set_1,
                                 const RowSet &row_set_2,
                                 RowSet *result)
{
	lock1(row_set_1);
	lock2(row_set_2);
	bool rc = LockedDifference::execute(result);
	unlock1();
	unlock2();
	return rc;
}

bool Difference(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
                  const char *column_name_1_0, ...)
{
	PreparedDifference prep_Difference;
	
	va_list ap;
	va_start(ap,column_name_1_0);
	const char *column_name_1 = column_name_1_0;
	while(column_name_1) {
		const char *column_name_2 = va_arg(ap,const char*);
		prep_Difference.add_relation(column_name_1_0,column_name_2);
		column_name_1 = va_arg(ap,const char*);
	}
	va_end(ap);
	return prep_Difference.execute(row_set_1,row_set_2,result);
}

