#include "ffdb_op.h"
#include "ffdb_op_cmpdigest.h"
#include <algorithm>
#include <stdarg.h>

#include <stdio.h>



bool LockedIntersection::prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2)
{
	if(cache_valid &&
	  row_set_1.descriptor()==cached_desc1 &&
	  row_set_2.descriptor()==cached_desc2)
		return true;
	cache_valid = false;
	
	if(!is_type_compatible(row_set_1.descriptor(),row_set_2.descriptor()))
		return false;
	cached_desc1 = row_set_1.descriptor();
	cached_desc2 = row_set_2.descriptor();
	if(!BinaryOpHelper::prepare_cache(cached_desc1,cached_desc2))
		return false;
	
	cache_valid = true;
	return true;
}


LockedIntersection::~LockedIntersection() {
}


bool LockedIntersection::execute(RowSet *result)
{
	if(!chunk1.rowset || !chunk2.rowset) return false;
	if(!prepare_cache(*chunk1.rowset,*chunk2.rowset))
		return false;
	process_rowset_1(false,false);
	process_rowset_2(true,false);
	
	result->setDescriptor(cached_desc1);
	
	Compare_RowDigest cmp;
	for(RowSetDigest::const_iterator i = chunk1.rowset_digest.begin();
	    i!=chunk1.rowset_digest.end();
	    i++)
	{
		RowSetDigest::const_iterator j = std::lower_bound(chunk2.rowset_digest.begin(),chunk2.rowset_digest.end(),*i,cmp);
		if(j!=chunk2.rowset_digest.end() && cmp.eq(*i,*j)) {
			result->push_back(*i->row_ptr);
		}
	}
	
	return true;
}



bool PreparedIntersection::execute(const RowSet &row_set_1,
                                   const RowSet &row_set_2,
                                   RowSet *result)
{
	lock1(row_set_1);
	lock2(row_set_2);
	bool rc = LockedIntersection::execute(result);
	unlock1();
	unlock2();
	return rc;
}


bool Intersection(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
                  const char *column_name_1_0, ...)
{
	PreparedIntersection prep_intersection;
	
	va_list ap;
	va_start(ap,column_name_1_0);
	const char *column_name_1 = column_name_1_0;
	while(column_name_1) {
		const char *column_name_2 = va_arg(ap,const char*);
		prep_intersection.add_relation(column_name_1_0,column_name_2);
		column_name_1 = va_arg(ap,const char*);
	}
	va_end(ap);
	return prep_intersection.execute(row_set_1,row_set_2,result);
}

