#include "ffdb_op.h"
#include "ffdb_op_alignment.h"
#include <stdarg.h>

#include <stdio.h>


void PreparedProjection::add_column(const char *column_name_) {
	column_name.push_back(column_name_);
	cached_result_desc.clear();
	cached_source_desc.clear();
}


bool PreparedProjection::build_cache(const RowSet &row_set) {
	const RowSetDescriptor &desc = row_set.descriptor();
	if(desc == cached_source_desc) return true;
	
	cached_source_desc.clear();
	cached_result_desc.clear();
	cached_result_desc.reserve(column_name.size());
	cached_col_idx.clear();
	cached_col_idx.reserve(column_name.size());
	
	for(unsigned i=0; i<column_name.size(); i++) {
		const std::string &cn=column_name[i];
		RowSetDescriptor::const_iterator j=desc.begin();
		unsigned col_idx=0;
		while(j!=desc.end()) {
			if(j->name==cn) break;
			j++;
			col_idx++;
		}
		if(j!=desc.end())
			cached_col_idx.push_back(col_idx);
		else
			return false;
	}
	
	cached_source_desc = desc;
	
	for(unsigned i=0; i<cached_col_idx.size(); i++) {
		unsigned col_idx = cached_col_idx[i];
		cached_result_desc.push_back(cached_source_desc[col_idx]);
	}
	
	return true;
}


bool PreparedProjection::execute(const RowSet &row_set, RowSet *result) {
	if(!build_cache(row_set)) return false;
	if(!result) return false;
	
	result->setDescriptor(cached_result_desc);
	result->clear();
	
	result->reserve(row_set.size());
	
	size_t size_hint=0;
	for(RowSet::const_iterator iter = row_set.begin();
	    iter!=row_set.end();
	    iter++)
	{
//printf("projecting row...\n");
		Row new_row;
		new_row.reserve_raw(size_hint);
		for(unsigned j=0; j<cached_col_idx.size(); j++) {
//printf("  projecting column...\n");
			unsigned col_idx=cached_col_idx[j];
			if(col_idx<iter->size()) {
				const Row::VL &vl = (*iter)[col_idx];
				ColumnType_t column_type = cached_result_desc[j].type;
				new_row.alloc(vl.size,ffdb_type_alignment[column_type],vl.ptr);
			} else
				new_row.alloc(0,0,0);
		}
		size_hint = new_row.raw_size();
		result->add_and_swap(new_row);
	}
	
	return true;
}





bool Projection(const RowSet &row_set, RowSet *result,
                const char *column_name0, ...)
{
	PreparedProjection proj;
	
	if(column_name0) {
		proj.add_column(column_name0);
		va_list ap;
		va_start(ap,column_name0);
		for(;;) {
			const char *column_name = va_arg(ap,const char*);
			if(!column_name) break;
			proj.add_column(column_name);
		}
		va_end(ap);
	}
	
	return proj.execute(row_set, result);
}

