#ifndef _FFDB_ROW_ACCESS_H_
#define _FFDB_ROW_ACCESS_H_

#include <assert.h>
#include "ffdb.h"
#include <inttypes.h>

inline RowSetDescriptor::size_type findDescriptorField(const RowSetDescriptor &desc, const char *name) {
#ifdef NDEBUG
	RowSetDescriptor::size_type i=0;
	while(desc[i].name!=name) i++;
	return i;
#else
	RowSetDescriptor::size_type i=0;
	for(;;) {
		assert(i<desc.size());
		if(desc[i].name==name) return i;
		i++;
	}
#endif
}


inline int64_t getInteger(RowSet::const_iterator iter, Row::size_type column_idx) {
	assert(column_idx<iter->size());
	assert((*iter)[column_idx].ptr);
	return *(const int64_t*)((*iter)[column_idx].ptr);
}

inline int64_t getInteger(const RowSetDescriptor &desc, RowSet::const_iterator iter, const char *name) {
	return getInteger(iter,findDescriptorField(desc,name));
}

inline int64_t getInteger(const RowSet &rowset, RowSet::const_iterator iter, const char *name) {
	return getInteger(iter,findDescriptorField(rowset.descriptor(),name));
}


inline double getReal(RowSet::const_iterator iter, Row::size_type column_idx) {
	assert(column_idx<iter->size());
	assert((*iter)[column_idx].ptr);
	return *(const double*)((*iter)[column_idx].ptr);
}

inline double getReal(const RowSetDescriptor &desc, RowSet::const_iterator iter, const char *name) {
	return getReal(iter,findDescriptorField(desc,name));
}

inline double getReal(const RowSet &rowset, RowSet::const_iterator iter, const char *name) {
	return getReal(iter,findDescriptorField(rowset.descriptor(),name));
}


inline const char *getString(RowSet::const_iterator iter, Row::size_type column_idx) {
	assert(column_idx<iter->size());
	assert((*iter)[column_idx].ptr);
	return (const char*)((*iter)[column_idx].ptr);
}

inline const char *getString(const RowSetDescriptor &desc, RowSet::const_iterator iter, const char *name) {
	return getString(iter,findDescriptorField(desc,name));
}

inline const char *getString(const RowSet &rowset, RowSet::const_iterator iter, const char *name) {
	return getString(iter,findDescriptorField(rowset.descriptor(),name));
}


inline const void *getRaw(RowSet::const_iterator iter, Row::size_type column_idx) {
	assert(column_idx<iter->size());
	assert((*iter)[column_idx].ptr);
	return (const void *)((*iter)[column_idx].ptr);
}

inline const void *getRaw(const RowSetDescriptor &desc, RowSet::const_iterator iter, const char *name) {
	return getRaw(iter,findDescriptorField(desc,name));
}

inline const void *getRaw(const RowSet &rowset, RowSet::const_iterator iter, const char *name) {
	return getRaw(iter,findDescriptorField(rowset.descriptor(),name));
}


inline bool getBool(RowSet::const_iterator iter, Row::size_type column_idx) {
	assert(column_idx<iter->size());
	assert((*iter)[column_idx].ptr);
	return *(const bool*)((*iter)[column_idx].ptr);
}

inline bool getBool(const RowSetDescriptor &desc, RowSet::const_iterator iter, const char *name) {
	return getBool(iter,findDescriptorField(desc,name));
}

inline bool getBool(const RowSet &rowset, RowSet::const_iterator iter, const char *name) {
	return getBool(iter,findDescriptorField(rowset.descriptor(),name));
}

#endif /*_FFDB_ROW_ACCESS_H_*/
