#include "ffdb.h"
#include "ffdb_convert.h"
#include "ffdb_io.h"
#include <stdio.h>
#include <string.h>
#include <inttypes.h>

int main(int argc, char **argv) {
	if(argc<3) {
		fprintf(stderr,"usage error\n");
		return -1;
	}
	RowSet source_rowset;
	if(!ReadFromFile(argv[1],&source_rowset)) {
		printf("Could not load DB\n");
		return -1;
	}
	const RowSetDescriptor &desc = source_rowset.descriptor();
	if(desc.size()!=4) {
		printf("4 columns expected in source DB\n");
		return -1;
	}
	
	RowSet result_rowset;
	result_rowset.setDescriptor(desc);
	for(RowSet::const_iterator iter=source_rowset.begin();
	    iter!=source_rowset.end();
	    iter++)
	{
		//first convert all columns to strings;
		char char_column0[128];
		bool char_column0_null;
		char char_column1[128];
		bool char_column1_null;
		char char_column2[128];
		bool char_column2_null;
		char char_column3[128];
		bool char_column3_null;
		RowConverter conv1;
		conv1.describe("a",char_column0,sizeof(char_column0),&char_column0_null);
		conv1.describe("b",char_column1,sizeof(char_column1),&char_column1_null);
		conv1.describe("c",char_column2,sizeof(char_column2),&char_column2_null);
		conv1.describe("d",char_column3,sizeof(char_column3),&char_column3_null);
		if(!conv1.toNative(desc,*iter)) {
			printf("Could not convert row\n");
			return -1;
		}
		
		//Then convert back into a row
		Row row1;
		if(!conv1.toRow(desc,&row1)) {
			printf("Could not convert back to row\n");
			return -1;
		}
		
		//then convert to reasonable native types
		char native_column0[128];
		bool native_column0_null;
		int native_column1;
		bool native_column1_null;
		bool native_column2;
		bool native_column2_null;
		double native_column3;
		bool native_column3_null;
		RowConverter conv2;
		conv2.describe("a",native_column0,sizeof(native_column0),&native_column0_null);
		conv2.describe("b",&native_column1,&native_column1_null);
		conv2.describe("c",&native_column2,&native_column2_null);
		conv2.describe("d",&native_column3,&native_column3_null);
		if(!conv2.toNative(desc,*iter)) {
			printf("Could not convert row\n");
			return -1;
		}
		
		//Then convert back into a row
		if(!conv2.toRow(desc,&row1)) {
			printf("Could not convert back to row\n");
			return -1;
		}
		
		result_rowset.push_back(row1);
	}
	
	if(!WriteToFile(argv[2],result_rowset)) {
		fprintf(stderr,"Could not write to test file\n");
		return -1;
	}
	return 0;
}
