#define __STDC_FORMAT_MACROS
#include "ffdb.h"
#include "ffdb_io.h"
#include "ffdb_op.h"
#include <stdio.h>
#include <string.h>
#include <inttypes.h>

static void decodehex(void * dst_, size_t /*dstsize*/, const char * src, size_t srcsize) {
	//to be optimized
	uint8_t * dst = (uint8_t*)dst_;
	while(srcsize) {
		char c1 = *src++;
		char c2 = *src++;
		srcsize-=2;
		uint8_t xv1;
		if(c1>='0' && c1<='9') xv1=(uint8_t)(c1-'0');
		else if(c1>='a' && c1<='f') xv1=(uint8_t)(c1-'a'+10);
		else xv1=(uint8_t)(c1-'A'+10);
		uint8_t xv2;
		if(c2>='0' && c2<='9') xv2=(uint8_t)(c2-'0');
		else if(c2>='a' && c2<='f') xv2=(uint8_t)(c2-'a'+10);
		else xv2=(uint8_t)(c2-'A'+10);
		uint8_t value = (uint8_t)((xv1<<4)|xv2);
		*dst++ = value;
	}
}

int main(int argc, char **argv) {
	if(argc<3) {
		fprintf(stderr,"usage error\n");
		return -1;
	}
	RowSet source_rowset;
	if(!ReadFromFile(argv[1],&source_rowset)) {
		printf("Could not load DB\n");
		return -1;
	}
	
	LockedSelection locked_sel;
	
	for(int i=3; i<argc; ) {
		const char *column_name=argv[i++];
		const char *cond_str=argv[i++];
		ConditionType_t condition_type;
		if(strcmp(cond_str,"null")==0)
			condition_type = cond_null;
		else if(strcmp(cond_str,"notnull")==0)
			condition_type = cond_notnull;
		else if(strcmp(cond_str,"lt")==0)
			condition_type = cond_lt;
		else if(strcmp(cond_str,"le")==0)
			condition_type = cond_le;
		else if(strcmp(cond_str,"eq")==0)
			condition_type = cond_eq;
		else if(strcmp(cond_str,"ge")==0)
			condition_type = cond_ge;
		else if(strcmp(cond_str,"gt")==0)
			condition_type = cond_gt;
		else if(strcmp(cond_str,"ne")==0)
			condition_type = cond_ne;
		if(condition_type!=cond_null && condition_type!=cond_notnull) {
			if(argv[i][0]!='c') {
				int cond_idx=locked_sel.add_condition(column_name,condition_type);
				if(cond_idx<0) {
					fprintf(stderr,"could not add condition %s %s\n", column_name, cond_str);
					return -1;
				}
				const char *value_str=argv[i++];
				bool set_rc;
				switch(*value_str) {
					case 'i': {
						int64_t value;
						sscanf(value_str+2,"%"SCNd64,&value);
						set_rc=locked_sel.set_value(cond_idx,value);
						break;
					}
					case 'f': {
						double value;
						sscanf(value_str+2,"%lf",&value);
						set_rc=locked_sel.set_value(cond_idx,value);
						break;
					}
					case 's': {
						const char *value = value_str+2;
						set_rc=locked_sel.set_value(cond_idx,value);
						break;
					}
					case 'r': {
						value_str+=2;
						size_t hl = strlen(value_str);
						char buf[128];
						decodehex(buf,hl/2,value_str,hl);
						set_rc=locked_sel.set_value(cond_idx,buf,hl/2);
						break;
					}
					case 'b': {
						bool value=false;
						if(strcmp(value_str+2,"true")==0)
							value=true;
						set_rc=locked_sel.set_value(cond_idx,value);
						break;
					}
					default: {
						fprintf(stderr,"usage error, could not understand '%s'\n",value_str);
						return -1;
					}
				}
				if(!set_rc) {
					fprintf(stderr,"Could not set value for %s %s\n", column_name, cond_str);
					return -1;
				}
			} else {
				const char *column_name2 = argv[i++]+2;
				int cond_idx=locked_sel.add_condition(column_name,condition_type,column_name2);
				if(cond_idx<0) {
					fprintf(stderr,"could nto add condition %s %s %s\n", column_name, cond_str, column_name2);
					return -1;
				}
			}
		} else {
			//null/notnull
			int cond_idx=locked_sel.add_condition(column_name,condition_type);
			if(cond_idx<0) {
				fprintf(stderr,"could not add condition %s %s\n", column_name, cond_str);
				return -1;
			}
		}
	}
	
	locked_sel.lock(source_rowset);
	
	RowSet result_rowset;
	if(!locked_sel.execute(&result_rowset)) {
		fprintf(stderr,"could not execute\n");
		return -1;
	}
	if(!locked_sel.execute(&result_rowset)) {
		fprintf(stderr,"could not execute\n");
		return -1;
	}
	
	if(!WriteToFile(argv[2],result_rowset)) {
		fprintf(stderr,"Could not write to test file\n");
		return -1;
	}
	return 0;
}
